/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.util.UniqueRealm;

/**
 * Introduces basic funtionality for ConcernJ implementation generators.
 *
 * @author Dennis Wagelaar
 * @version 0.1.21, 6 April 2004
 */

public abstract class ConcernJImplementationGenerator
  extends JavaImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ConcernJImplementationGenerator.
   */
  public ConcernJImplementationGenerator()
  {
  } //ConcernJImplementationGenerator

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /* Serialises a list of declarations into an indented string. */
  private String serialiseDecls(Collection decls)
  {
    String result = new String();
    Iterator items = decls.iterator();
    String item;

    while (items.hasNext())
    {
      item = (String) items.next();
      result += "      " + item + "\n";
    }
    return result;
  } //serialiseDecls

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the name to use for a filtermodule based on the composite's name.
   * @return The filtermodule name.
   */
  protected String getFiltermoduleName()
  {
    return (lcfirst(getComposite().getName()) + "Module");
  } //getFiltermoduleName

  /**
   * Gets the implementation of the unique realm. Creates one if none-existent.
   * @param realm The unique realm.
   * @return The realm implementation.
   */
  protected String getRealmImplementation(UniqueRealm realm)
  {
    String impl = getVisitor().getImpl(realm);
    if (impl == null)
    {
      impl = "concern " + realm.getName() + " begin\n" +
             "  implementation in java file \"" + realm.getName() + ".jwf\";\n" +
             "end concern " + realm.getName() + ";\n" +
             "<JAVA>\n" + // separator to create .cf and .jwf files
             "public class " + realm.getName() + " {\n}";
    }
    return impl;
  } //getRealmImplementation

  /**
   * Inserts parameters into the CF and Java method declaration and returns the
   * updated implementation.
   *
   * @param impl The current implementation.
   * @param parms The parameter declarations. Structure: <name> |-> <type>.
   * @return The updated implementation.
   */
  protected String insertParameters(String impl, Hashtable parms)
  {
    String result, parm, type;
    int startIndex;
    Enumeration ps;
    boolean first = true;

    // weave new parameters
    startIndex = impl.indexOf("(") + 1;
    result = impl.substring(0, startIndex);
    ps = parms.keys();
    while (ps.hasMoreElements())
    {
      if (!first)
        result += ", ";
      parm = (String) ps.nextElement();
      type = (String) parms.get(parm);
      result += parm + " : " + type;
      first = false;
    }
    result += super.insertParameters(impl.substring(startIndex), parms);
    return result;
  } //insertParameters

  /**
   * Inserts a return type clause into the implementation and returns the updated
   * implementation.
   *
   * @param impl The current implementation.
   * @param type The return type.
   * @return The updated implementation.
   */
  protected String insertReturnType(String impl, String type)
  {
    int insertIndex;

    // Java return type
    impl = super.insertReturnType(impl, type);
    // get insertion point (delete previous return type (void?))
    insertIndex = impl.indexOf(");") + 1;
    // insert and return
    impl = impl.substring(0, insertIndex) + " : " + type +
      impl.substring(insertIndex);
    return impl;
  } //insertReturnType

  /**
   * Inserts a method override id into the implementation and returns the
   * updated implementation.
   *
   * @param impl The current implementation.
   * @param id The overridden method id.
   * @return The updated implementation.
   */
  protected String insertMethodOverrideID(String impl, int id)
  {
    String j_impl = impl.substring(impl.indexOf("\n<JAVA>\n") + 8);
    // Java override ID
    j_impl = super.insertMethodOverrideID(j_impl, id);
    // insert (overwrite) and return
    impl = "<" + id + ".cf_header>\n<JAVA>\n" + j_impl;
    return impl;
  } //insertMethodOverrideID

  /**
   * Creates a filtermodule block into the given implementation and returns it.
   *
   * @param moduleName The name of the filtermodule.
   * @param internals The internal declarations.
   * @param externals The external declarations.
   * @param conditions The condition declarations.
   * @param methods The method declarations.
   * @param inputfilters The inputfilter declarations.
   * @param outputfilters The outputfilter declarations.
   * @return The created filtermodule.
   */
  protected String createFiltermodule(String moduleName,
    Collection internals, Collection externals, Collection conditions,
    Collection methods, Collection inputfilters, Collection outputfilters)
  {
    String filterImpl;
    boolean empty = true;

    filterImpl = "  filtermodule " + moduleName + " begin\n";
    // insert internals
    if (internals != null)
    {
      filterImpl += "    internals\n";
      filterImpl += serialiseDecls(internals);
      empty = false;
    }
    // insert externals
    if (externals != null)
    {
      filterImpl += "    externals\n";
      filterImpl += serialiseDecls(externals);
      empty = false;
    }
    // insert conditions
    if (conditions != null)
    {
      filterImpl += "    conditions\n";
      filterImpl += serialiseDecls(conditions);
      empty = false;
    }
    // insert methods
    if (methods != null)
    {
      filterImpl += "    methods\n";
      filterImpl += serialiseDecls(methods);
      empty = false;
    }
    // insert inputfilters
    if (inputfilters != null)
    {
      filterImpl += "    inputfilters\n";
      filterImpl += serialiseDecls(inputfilters);
      empty = false;
    }
    // insert outputfilters
    if (outputfilters != null)
    {
      filterImpl += "    outputfilters\n";
      filterImpl += serialiseDecls(outputfilters);
      empty = false;
    }
    filterImpl += "  end filtermodule " + moduleName + ";";
    if (empty)
      return new String();
    else
      return filterImpl;
  } //createFiltermodule

  /**
   * Inserts a filtermodule block into the given implementation and returns
   * the updated implementation.
   *
   * @param impl The current implementation.
   * @param filterModule The filtermodule implementation.
   * @return The updated implementation.
   */
  protected String insertFiltermodule(String impl, String filterModule)
  {
    int filterIndex;

    // insert filtermodule
    if (filterModule.length() > 0)
    {
      filterIndex = impl.indexOf('\n');
      impl = impl.substring(0, filterIndex + 1) + filterModule + '\n' +
        impl.substring(filterIndex + 1);
    }
    return impl;
  } //insertFiltermodule

  /**
   * Inserts a superimposition block into the given implementation and returns
   * the updated implementation.
   *
   * @param impl The current implementation.
   * @param selectors The selector declarations.
   * @param conditions The condition declarations.
   * @param methods The method declarations.
   * @param filtermodules The filtermodule declarations.
   * @return The updated implementation.
   */
  protected String insertSuperimposition(String impl, Collection selectors,
    Collection conditions, Collection methods, Collection filtermodules)
  {
    Iterator items;
    int minIndex, startIndex, endIndex;

    // weave superimposition block with current implementation
    if ((minIndex = impl.indexOf("superimposition begin")) == -1)
    {
      minIndex = impl.indexOf("  implementation in");
      impl = impl.substring(0, minIndex) +
             "  superimposition begin\n" +
             "  end superimposition;\n" +
             impl.substring(minIndex);
    }
    // weave selectors
    if (selectors != null)
    {
      if ((startIndex = impl.indexOf("    selectors", minIndex)) == -1)
      {
        if ((startIndex = impl.indexOf("    conditions", minIndex)) == -1)
          if ((startIndex = impl.indexOf("    methods", minIndex)) == -1)
            if ((startIndex = impl.indexOf("    filtermodules", minIndex)) == -1)
              startIndex = impl.indexOf("  end superimposition", minIndex);
        endIndex = startIndex;
      }
      else
        endIndex = startIndex + 14;
      impl = impl.substring(0, startIndex) +
             "    selectors\n" + serialiseDecls(selectors) +
             impl.substring(endIndex);
    }
    // weave conditions
    if (conditions != null)
    {
      if ((startIndex = impl.indexOf("    conditions", minIndex)) == -1)
      {
        if ((startIndex = impl.indexOf("    methods", minIndex)) == -1)
          if ((startIndex = impl.indexOf("    filtermodules", minIndex)) == -1)
            startIndex = impl.indexOf("  end superimposition", minIndex);
        endIndex = startIndex;
      }
      else
        endIndex = startIndex + 12;
      impl = impl.substring(0, startIndex) +
             "    conditions\n" + serialiseDecls(conditions) +
             impl.substring(endIndex);
    }
    // weave methods
    if (methods != null)
    {
      if ((startIndex = impl.indexOf("    methods", minIndex)) == -1)
      {
        if ((startIndex = impl.indexOf("    filtermodules", minIndex)) == -1)
          startIndex = impl.indexOf("  end superimposition", minIndex);
        endIndex = startIndex;
      }
      else
        endIndex = startIndex + 15;
      impl = impl.substring(0, startIndex) +
             "    methods\n" + serialiseDecls(methods) +
             impl.substring(endIndex);
    }
    // weave filtermodules
    if (filtermodules != null)
    {
      if ((startIndex = impl.indexOf("    filtermodules", minIndex)) == -1)
      {
        startIndex = impl.indexOf("  end superimposition", minIndex);
        endIndex = startIndex;
      }
      else
        endIndex = startIndex + 18;
      impl = impl.substring(0, startIndex) +
             "    filtermodules\n" + serialiseDecls(filtermodules) +
             impl.substring(endIndex);
    }

    return impl;
  } //insertSuperimposition

  /**
   * Returns a filter declaration given the filter name, type and rules.
   *
   * @param name The filter name.
   * @param type The filter type.
   * @param rules The set of filter rules.
   * @return The filter declaration implementation.
   */
  protected String makeFilterDecl(String name, String type, Collection rules)
  {
    String filter;
    Iterator ruleIt;

    filter = name + " : " + type + " = {\n";
    ruleIt = rules.iterator();
    while(ruleIt.hasNext())
    {
      filter += "        " + (String) ruleIt.next() + "\n";
    }
    filter += "      };";
    return filter;
  } //makeFilterDecl
} //class ConcernJImplementationGenerator
