/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing an Inheritance composite
 * in ConcernJ.
 *
 * @author Dennis Wagelaar
 * @version 0.1.06, 7 Maart 2003
 */

public class ConcernJInheritance extends ConcernJImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ConcernJInheritance
   */
  public ConcernJInheritance()
  {
  } //ConcernJInheritance

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the parent role of the composite.
   * @return The parent CompositeRole.
   */
  protected CompositeRole getParentRole()
  {
    return getComposite().getRole("Parent");
  } //getParentRole

  /**
   * Retrieves the child role of the composite.
   * @return The child CompositeRole.
   */
  protected CompositeRole getChildRole()
  {
    return getComposite().getRole("Child");
  } //getChildRole

  /**
   * Checks the composite roles for existence.
   * @throws CoComposeException if a required role does not exist.
   */
  protected void checkRoles() throws CoComposeException
  {
    Composite composite = getComposite();
    if (getParentRole() == null)
      throw new CoComposeException("Parent role not found in composite " +
        composite.getName(), composite);
    if (getChildRole() == null)
      throw new CoComposeException("Child role not found in composite " +
        composite.getName(), composite);
  } //checkRoles

  /**
   * Creates an internal declaration from the given concept.
   * @param concept The concept to create the internal declaration for.
   * @return The internal declaration.
   */
  protected String createInternal(Concept concept)
  {
    return (getInstanceName(concept) + " : " + concept.getName() + ";");
  } //createInternal

  /**
   * Creates a filter rule from the given concept.
   * @param concept The concept to create the filter rule for.
   * @return The filter rule.
   */
  protected String createFilterRule(Concept concept)
  {
    return ("true => " + getInstanceName(concept) + ".*;");
  } //createFilterRule

  /**
   * Creates a collection of internal declarations given a collection of class
   * concepts.<BR>
   * PRE: classes ::= Collection&lt;Concept&gt;<BR>
   * POST: result ::= Collection&lt;String&gt;
   * @param classes The collection of class concepts.
   * @return The collection of internal declarations.
   * @throws CoComposeException
   */
  protected Collection createInternals(Collection classes)
    throws CoComposeException
  {
    Collection internals = new Vector();
    Concept concept;
    Iterator classIt = classes.iterator();
    try
    {
      while (classIt.hasNext())
      {
        concept = (Concept) classIt.next();
        internals.add(createInternal(concept));
      } //while
    } //try
    catch (ClassCastException e)
    {
      throw new CoComposeException(getClass().getName() + ".createInternals" +
        "(): List of primitives contains non-Concept: " +
        e.getMessage(), getComposite());
    } //catch
    return internals;
  } //createInternals

  /**
   * Creates a collection of filter rules given a collection of class
   * concepts.<BR>
   * PRE: parents ::= Collection&lt;Concept&gt;<BR>
   * POST: result ::= Collection&lt;String&gt;
   * @param parents The collection of parent class concepts.
   * @return The collection of filter rules.
   * @throws CoComposeException
   */
  protected Collection createFilterRules(Collection parents)
    throws CoComposeException
  {
    Collection filterrules = new Vector();
    Concept concept;
    Iterator parentIt = parents.iterator();
    try
    {
      while (parentIt.hasNext())
      {
        concept = (Concept) parentIt.next();
        filterrules.add(createFilterRule(concept));
      } //while
    } //try
    catch (ClassCastException e)
    {
      throw new CoComposeException(getClass().getName() + ".createFilterRules" +
        "(): List of primitives contains non-Concept: " +
        e.getMessage(), getComposite());
    } //catch
    return filterrules;
  } //createFilterRules

  /**
   * Creates the input filters for inheritance.<BR>
   * PRE: parents ::= Collection&lt;Concept&gt;<BR>
   * POST: result ::= Collection&lt;String&gt;
   * @param parents The parent class concepts.
   * @return The collection of input filter declarations
   * @throws CoComposeException
   */
  protected Collection createInputFilters(Collection parents)
    throws CoComposeException
  {
    Collection inputfilters = new Vector();
    inputfilters.add(makeFilterDecl("inherits", "Dispatch",
      createFilterRules(parents)));
    return inputfilters;
  } //createInputFilters

  /**
   * Creates the filtermodules declarations for use in the superimposition
   * block.<BR>
   * POST: result ::= Collection&lt;String&gt;
   * @param target The target identifier of the superimposition.
   * @return The collection of filtermodule declarations.
   */
  protected Collection createFiltermodulesDecl(String target)
  {
    Collection filtermodules = new Vector();
    filtermodules.add(target + " <- { " + getFiltermoduleName() + " };");
    return filtermodules;
  } //createFiltermodulesDecl

  /**
   * Returns the selector name based on the composite.
   * @return The selector name.
   */
  protected String getSelectorName()
  {
    return getInstanceName(getComposite());
  } //getSelectorName

  /**
   * Creates a selector declaration for use in the
   * superimposition block.<BR>
   * PRE: concepts ::= Collection&lt;Concept&gt;
   * @param concepts The collection of concepts to put in the selector.
   * @return The selector declaration.
   */
  protected String createSelector(Collection concepts)
  {
    Concept concept;
    String selector = getSelectorName() + " = Set(";
    Iterator conceptIt = concepts.iterator();
    while (conceptIt.hasNext())
    {
      concept = (Concept) conceptIt.next();
      selector += concept.getName() + ", ";
    }
    selector = selector.substring(0, selector.lastIndexOf(",")) + ");";
    return selector;
  } //createSelector

  /**
   * Creates a collection of selector declarations for use in the
   * superimposition block.<BR>
   * PRE: concepts ::= Collection&lt;Concept&gt;<BR>
   * POST: result ::= Collection&lt;String&gt;
   * @param concepts The collection of concepts to put in the selector.
   * @return The collection of selector declarations.
   */
  protected Collection createSelectors(Collection concepts)
  {
    Collection selectors = new Vector();
    selectors.add(createSelector(concepts));
    return selectors;
  } //createSelectors

  /**
   * Applies the implementation generator for each Child concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyPerChild() throws CoComposeException
  {
    Object link;
    Collection parents = getConceptsFor(getParentRole());
    Iterator links = getModel().getLinkedNodes(getChildRole()).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof Concept)
        applyToChild((Concept) link, parents);
    } //while
  } //applyPerChild

  /**
   * Applies the implementation generator to a Child concept.
   * @param child The concept to apply the role part for.
   * @param parents The set of parent concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToChild(Concept child, Collection parents)
    throws CoComposeException
  {
    String impl;

    // get current implementation
    impl = getVisitor().getImpl(child);
    impl = insertPrimitiveImplementations(impl, getPrimitives(parents));
    impl = insertFiltermodule(impl, createFiltermodule(getFiltermoduleName(),
      createInternals(parents), null, null, null,
      createInputFilters(parents), null));
    impl = insertSuperimposition(impl, null, null, null,
      createFiltermodulesDecl("self"));
    // set implementation
    getVisitor().setImpl(child, impl);
  } //applyToChild

  /**
   * Applies the implementation generator to a UniqueRealm.
   * @param realm The unique realm.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToRealm(UniqueRealm realm)
    throws CoComposeException
  {
    String impl;
    Collection children = getConceptsFor(getChildRole());
    Collection parents = getConceptsFor(getParentRole());

    // if no children, no implementation needed
    if (children.size() == 0)
      return;
    impl = getRealmImplementation(realm);
    impl = insertPrimitiveImplementations(impl, getPrimitives(parents));
    impl = insertFiltermodule(impl, createFiltermodule(getFiltermoduleName(),
      createInternals(parents), null, null, null,
      createInputFilters(parents), null));
    impl = insertSuperimposition(impl, createSelectors(children), null, null,
      createFiltermodulesDecl(getSelectorName()));
    // set implementation
    getVisitor().setImpl(realm, impl);
  } //applyToRealm

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    UniqueRealm realm;

    super.apply(implementation, generateImpl);
    realm = getComposite().getUniqueRealm();
    checkRoles();
    // check whether or not to use a UniqueRealm for implementation
    if (realm == getModel())
    { // no special realm, so apply per child concept
      applyPerChild();
    }
    else
    { // retrieve children and apply for UniqueRealm
      applyToRealm(realm);
    }
  } //apply
} //class ConcernJInheritance
