/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;

import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing a SingleRelation composite
 * in ConcernJ.
 *
 * @author Dennis Wagelaar
 * @version 0.1.21, 6 April 2003
 */

public class ConcernJSingleRelation extends ConcernJImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ConcernJSingleRelation
   */
  public ConcernJSingleRelation()
  {
  } //ConcernJSingleRelation

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Checks whether any of the Property concepts are overriding by means of an
   * Overrides composite.
   * @param properties the Property concepts.
   * @return true if any of the Property concepts are overriding.
   */
  private boolean isOverriding(Collection properties)
  {
    /* THIS is dirty hardcoding b.t.w... This implementation generator requires
     * intricate knowledge about the presence of the "Overrides" composite type.
     * Never mind the level of nested loops b.t.w...
     */ 
    Model model = getComposite().getModel();
    UniqueRealm realm = getComposite().getUniqueRealm();
    Iterator props = properties.iterator();
    while (props.hasNext())
    {
      Concept property = (Concept) props.next();
      for (int ch = property.firstChild(); ch != -1; ch = property.nextChild())
      { // check what the property is linked to
        Node node = model.getNodeFromIndex(ch);
        if (node instanceof CompositeRole)
        { // linked to a CompositeRole
          CompositeRole role = (CompositeRole) node;
          Composite composite = role.getComposite();
          if (role.getName().equals("NewMethod") && 
              composite.getType().equals("Overrides"))
          { // BINGO: this property concept is overriding another concept,
            // which will "steal" it's name => possible namespace problems for our
            // ConcernJ concern.
            return true;
          }
        }
      } //for
    } //while
    return false;
  }

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator to an Owner concept.
   *
   * @param owner The concept to apply the role part for.
   * @param properties The set of property concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwner(Concept owner, Collection properties)
    throws CoComposeException
  {
    Composite composite = getComposite();
    Collection externals = new Vector();
    Collection conditions = new Vector();
    Collection conditionSuper = new Vector();
    Collection methods = new Vector();
    Collection methodSuper = new Vector();
    Collection inputfilters = new Vector();
    Collection outputfilters = new Vector();
    Collection filtermodules = new Vector();
    Iterator propertyIt;
    Concept property;
    String propertyName, fmodName, impl, condSuper, methSuper;
    int propertyCount;
    Hashtable parms = new Hashtable();
    String form = owner.getForm();

    propertyCount = properties.size();
    // if no properties, no implementation needed
    if (propertyCount == 0)
      return;
    impl = getCurrentImplementation(owner);
    // generate filtermodule and superimposition declarations
    fmodName = lcfirst(composite.getName()) + "Module";
    condSuper = "self <- { ";
    methSuper = "self <- { ";
    propertyIt = properties.iterator();
    while (propertyIt.hasNext())
    {
      property = (Concept) propertyIt.next();
      propertyName = property.getName();
      form = property.getForm();
      if (form.equals("Filtermodule"))
      { // add filtermodule implementation and prepare for superimposition
        impl = insertFiltermodule(impl, "  <" + property.getId() + ">");
        filtermodules.add("self <- { <" + property.getId() + ".cf_name> };");
      }
      else if (form.equals("Condition"))
      {
        // add pointer to property, since property implementation may not be
        // finished.
        conditions.add("<" + property.getId() + ".cf_header>");
        condSuper += "<" + property.getId() + ".cf_name>, ";
        impl = insertMethod(impl, "<" + property.getId() + ">");
      }
      else if (form.equals("Method"))
      {
        // add pointer to property, since property implementation may not be
        // finished.
        methods.add("<" + property.getId() + ".cf_header>");
        methSuper += "<" + property.getId() + ".cf_name>, ";
        impl = insertMethod(impl, "<" + property.getId() + ">");
      }
      else
        throw new CoComposeException("No Property role part defined for concept form "
          + form, property);
    }
    // check inputfilters
    if (inputfilters.isEmpty())
      inputfilters = null;
    // check outputfilters
    if (outputfilters.isEmpty())
      outputfilters = null;
    // check externals
    if (externals.isEmpty())
      externals = null;
    // check conditions
    if (conditions.isEmpty())
    {
      conditions = null;
      conditionSuper = null;
    }
    else
    {
      condSuper = condSuper.substring(0, condSuper.lastIndexOf(",")) + " };";
      conditionSuper.add(condSuper);
    }
    // check methods
    if (methods.isEmpty())
    {
      methods = null;
      methodSuper = null;
    }
    else
    {
      methSuper = methSuper.substring(0, methSuper.lastIndexOf(",")) + " };";
      methodSuper.add(methSuper);
    }
    // insert filtermodule
    impl = insertFiltermodule(impl, createFiltermodule(fmodName, null,
      externals, conditions, methods, inputfilters, outputfilters));
    // insert superimposition
    filtermodules.add("self <- { " + fmodName + " };");
    impl = insertSuperimposition(impl, null, conditionSuper, methodSuper,
      filtermodules);
    // set implementation
    setCurrentImplementation(owner, impl);
  } //applyToOwner

  /**
   * Applies the implementation generator to a UniqueRealm.
   *
   * @param realm The unique realm.
   * @param properties The set of property concepts.
   * @param owners The set of owner concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToRealm(UniqueRealm realm, Collection properties, Collection owners)
    throws CoComposeException
  {
    Concept owner;
    Composite composite = getComposite();
    Collection externals = new Vector();
    Collection conditions = new Vector();
    Collection conditionSuper = new Vector();
    Collection methods = new Vector();
    Collection methodSuper = new Vector();
    Collection inputfilters = new Vector();
    Collection outputfilters = new Vector();
    Collection filtermodules = new Vector();
    Collection selectors = new Vector();
    Iterator ownerIt, propertyIt;
    Concept property;
    String selName, fmodName, impl, form, selector, condSuper, methSuper;
    int propertyCount, ownerConcernCount = 0;
    Hashtable parms = new Hashtable();

    propertyCount = properties.size();
    // if no properties, no implementation needed
    if (propertyCount == 0)
      return;
    // init
    selName = lcfirst(composite.getName());
    fmodName = selName + "Module";
    selector = selName + " = Set(";
    condSuper = "self <- { ";
    methSuper = selName + " <- { ";
    // divide owners in Methods and Concerns
    ownerIt = owners.iterator();
    while (ownerIt.hasNext())
    {
      owner = (Concept) ownerIt.next();
      form = owner.getForm();
      if (form.equals("Concern"))
      { // add directly to selector
        selector += owner.getName() + ", ";
        ownerConcernCount++;
      }
      else
        throw new CoComposeException("No Owner role part defined for concept form "
          + form, owner);
    }
    // stop if no owner concerns
    if (ownerConcernCount == 0)
      return;
    selector = selector.substring(0, selector.lastIndexOf(",")) + ");";
    // concern owner role parts are put in a realm concern
    impl = getRealmImplementation(realm);
    // generate filtermodule and superimposition declarations
    propertyIt = properties.iterator();
    while (propertyIt.hasNext())
    {
      property = (Concept) propertyIt.next();
      form = property.getForm();
      if (form.equals("Filtermodule"))
      { // add filtermodule implementation and prepare for superimposition
        impl = insertFiltermodule(impl, "  <" + property.getId() + ">");
        filtermodules.add(selName + " <- { <" + property.getId() + ".cf_name> };");
      }
      else if (form.equals("Condition"))
      {
        // add pointer to property, since property implementation may not be
        // finished.
        conditions.add("<" + property.getId() + ".cf_header>");
        condSuper += "<" + property.getId() + ".cf_name>, ";
        impl = insertMethod(impl, "<" + property.getId() + ">");
      }
      else if (form.equals("Method"))
      {
        // add pointer to property, since property implementation may not be
        // finished.
        methods.add("<" + property.getId() + ".cf_header>");
        methSuper += "<" + property.getId() + ".cf_name>, ";
        impl = insertMethod(impl, "<" + property.getId() + ">");
      }
      else
        throw new CoComposeException("No Property role part defined for concept form "
          + form, property);
    }
    // check inputfilters
    if (inputfilters.isEmpty())
      inputfilters = null;
    // check outputfilters
    if (outputfilters.isEmpty())
      outputfilters = null;
    // check externals
    if (externals.isEmpty())
      externals = null;
    // check conditions
    if (conditions.isEmpty())
    {
      conditions = null;
      conditionSuper = null;
    }
    else
    {
      condSuper = condSuper.substring(0, condSuper.lastIndexOf(",")) + " };";
      conditionSuper.add(condSuper);
    }
    // check methods
    if (methods.isEmpty())
    {
      methods = null;
      methodSuper = null;
    }
    else
    {
      methSuper = methSuper.substring(0, methSuper.lastIndexOf(",")) + " };";
      methodSuper.add(methSuper);
    }
    // insert filtermodule
    impl = insertFiltermodule(impl, createFiltermodule(fmodName, null,
      externals, conditions, methods, inputfilters, outputfilters));
    // insert superimposition
    selectors.add(selector);
    filtermodules.add(selName + " <- { " + fmodName + " };");
    impl = insertSuperimposition(impl, selectors, conditionSuper, methodSuper,
      filtermodules);
    // set implementation
    setCurrentImplementation(realm, impl);
  } //applyToRealm

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    Collection owners = new Vector();
    Collection properties = new Vector();
    Composite composite = implementation.getComposite();
    CompositeRole ownerRole = composite.getRole("Owner");
    CompositeRole propertyRole = composite.getRole("Property");
    UniqueRealm realm = composite.getUniqueRealm();
    Model model = composite.getModel();
    Iterator links;
    Object link;

    // superclass apply
    super.apply(implementation, generateImpl);
    // check roles
    if (ownerRole == null)
      throw new CoComposeException("Owner role not found in composite " +
        composite.getName(), composite);
    if (propertyRole == null)
      throw new CoComposeException("Property role not found in composite " +
        composite.getName(), composite);
    // retrieve properties
    links = model.getLinkedNodes(propertyRole).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof EDU.utwente.CoCompose.lang.Concept)
        properties.add(link);
    }
    // check whether or not to use a UniqueRealm for implementation
    if ((realm == model) || (isOverriding(properties)))
    { // no special realm, so apply per owner concept
      links = model.getLinkedNodes(ownerRole).iterator();
      while (links.hasNext())
      {
        link = links.next();
        if (link instanceof EDU.utwente.CoCompose.lang.Concept)
          applyToOwner((Concept) link, properties);
      } //while
    }
    else
    { // retrieve owners and apply for UniqueRealm
      links = model.getLinkedNodes(ownerRole).iterator();
      while (links.hasNext())
      {
        link = links.next();
        if (link instanceof EDU.utwente.CoCompose.lang.Concept)
          owners.add(link);
      } //while
      applyToRealm(realm, properties, owners);
    }
  } //apply
} //ConcernJSingleRelation
