/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.Collection;
import java.util.Vector;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing a SingleInstance
 * composite in the JAsCo language.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 22 November 2003
 */

public class JAsCoSingleInstance extends JavaSingleInstance
{
  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Finds the AspectBean concept for a Hook instance.
   * @param propertyImpl the property (instance) implementation
   * @return the corresponding AspectBean concept name including dot (e.g. "Aspect.") 
   */
  private String findAspect(String propertyImpl)
  {
    int startIndex, endIndex;
    if ((endIndex = propertyImpl.indexOf('.')) != -1)
    {
      if ((startIndex = propertyImpl.lastIndexOf(' ', endIndex)) != -1)
      {
        return propertyImpl.substring(startIndex + 1, endIndex + 1);
      }
    }
    return "";
  } //findAspect

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator to an Owner concept.
   * @param owner The concept to apply the role part for.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwner(Concept owner)
    throws CoComposeException
  {
    Concept propertyC = (Concept) getConceptsFor(getPropertyRole()).iterator().next();
    Concept classC = (Concept) getConceptsFor(getClassRole()).iterator().next();
    String form = owner.getForm();
    // Dispatch by owner concept form
    if (form.endsWith("Connector"))
      applyToOwnerConnector(owner, propertyC, classC);
    else
      super.applyToOwner(owner);
  } //applyToOwner

  /**
   * Applies the implementation generator to an Owner concept with form Connector.
   * @param owner The concept to apply the role part for.
   * @param propertyC The property concept.
   * @param classC The class concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwnerConnector(Concept owner, Concept propertyC,
    Concept classC) throws CoComposeException
  {
    Collection variables = new Vector();
    String impl, variable;

    impl = getCurrentImplementation(owner);
    // add Primitive implementation tag, if necessary
    if (isPrimitive(classC.getForm()))
      impl = insertPrimitiveImplementation(impl, classC);
    // generate variables
    variable = findAspect(getCurrentImplementation(propertyC)) + 
      classC.getName() + " " + propertyC.getName() + "<" + 
      propertyC.getId() + ">;";
    variables.add(variable);
    if (!variables.isEmpty())
      impl = insertVariables(impl, variables);
    setCurrentImplementation(owner, impl);
  } //applyToOwnerConnector
} //class JAsCoSingleInstance