/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing a SingleRelation
 * composite in Java.
 *
 * @author Dennis Wagelaar
 * @version 0.1.06, 5 February 2003
 */

public class JavaSingleRelation extends JavaImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a JavaSingleRelation.
   */
  public JavaSingleRelation()
  {
  } //JavaSingleRelation

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the Owner role of the composite.
   * @return The Owner CompositeRole.
   */
  protected CompositeRole getOwnerRole()
  {
    return getComposite().getRole("Owner");
  } //getOwnerRole

  /**
   * Retrieves the Property role of the composite.
   * @return The Property CompositeRole.
   */
  protected CompositeRole getPropertyRole()
  {
    return getComposite().getRole("Property");
  } //getPropertyRole

  /**
   * Checks the composite roles for existence.
   * @throws CoComposeException if a required role does not exist.
   */
  protected void checkRoles() throws CoComposeException
  {
    Composite composite = getComposite();
    if (getOwnerRole() == null)
      throw new CoComposeException("Owner role not found in composite " +
        composite.getName(), composite);
    if (getPropertyRole() == null)
      throw new CoComposeException("Property role not found in composite " +
        composite.getName(), composite);
  } //checkRoles

  /**
   * Applies the implementation generator for each Owner concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyPerOwner() throws CoComposeException
  {
    Object link;
    Iterator links = getModel().getLinkedNodes(getOwnerRole()).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof Concept)
        applyToOwner((Concept) link, getConceptsFor(getPropertyRole()));
    } //while
  } //applyPerOwner

  /**
   * Applies the implementation generator to an Owner concept.
   *
   * @param owner The concept to apply the role part for.
   * @param properties The set of property concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwner(Concept owner, Collection properties)
    throws CoComposeException
  {
    Iterator propertyIt;
    Concept property;
    String impl;

    impl = getCurrentImplementation(owner);
    // generate methods
    propertyIt = properties.iterator();
    while (propertyIt.hasNext())
    {
      property = (Concept) propertyIt.next();
      // add pointer to property, since property implementation may not be
      // finished.
      impl = insertMethod(impl, "<" + property.getId() + ">");
    }
    setCurrentImplementation(owner, impl);
  } //applyToOwner

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    super.apply(implementation, generateImpl);
    checkRoles();
    applyPerOwner();
  } //apply
} //JavaSingleRelation
